﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Collections;

namespace NWNHeatMap
{
    class Gradient
    {
        List<GradientPoint> points;

        public Gradient(Color start, Color end)
        {
            points = new List<GradientPoint>();
            AddPoint(start, 0.0f);
            AddPoint(end, 1.0f);
        }

        public void AddPoint(Color col, float pos)
        {
            points.Add(new GradientPoint(col, pos));
            points.Sort(new Comparison<GradientPoint>(GPSortFunc));
        }

        private int GPSortFunc(GradientPoint gp1, GradientPoint gp2)
        {
            if (gp1.Position - gp2.Position < 0) return -1;
            else if (gp1.Position - gp2.Position > 0) return 1;
            else return 0;
        }

        public Color MapPosToColor(float pos)
        {
            // Special case for zero position
            if (pos == 0.0f) return points[0].Color;
            // Step through the points in the gradient
            for (int i = 0; i < points.Count; i++)
            {
                // Is pos greater than the last one but <= the next one?
                if (points[i].Position < pos
                    && points[i+1].Position >= pos) // should be no need to check array bounds since pos should never be >1.0f
                {
                    if (i == points[i+1].Position) return points[i+1].Color;
                    else // we're not perfectly on a point, so we must lerp
                    {
                        float weight = (pos - points[i].Position) / (points[i + 1].Position - points[i].Position);
                        return HeatMap.Lerp(points[i].Color, points[i+1].Color, weight);
                    }
                }
            }
            throw new Exception("Couldn't find position in gradient");
        }

        /// <summary>
        /// Finds evenly distributed percentiles normalized for use in gradients
        /// </summary>
        /// <param name="points">Sorted list of points</param>
        /// <param name="divisions">Number of percentiles to find (excludes extrema, since they're trivial)</param>
        /// <returns></returns>
        public static List<float> GetDivisionsFromData(List<float> points, int divisions)
        {
            if (divisions >= points.Count)
                throw new Exception("Insufficient data points for number of divisions requested");

            float min_intensity = points[0];
            float max_intensity = points[points.Count - 1];

            // Throw away all 0-intensity points; they just skew the divisions down
            List<float> newpoints = new List<float>();
            foreach (float f in points)
            {
                if(f > 0.0f)
                    newpoints.Add(f);
            }

            List<float> divs = new List<float>();

            for (int i = 0; i < divisions; i++)
            {
                float pctl_idx = (i + 1) * ((float)newpoints.Count / (divisions + 1));
                float weight = pctl_idx - (int)pctl_idx;
                float lower = newpoints[(int)pctl_idx];
                float upper = newpoints[(int)pctl_idx + 1];
                float pctl_val = lower * (1 - weight) + upper * weight;
                float pctl_norm = pctl_val / (max_intensity - min_intensity);
                divs.Add(pctl_norm);
            }

            return divs;
        }
    }

    class GradientPoint
    {
        Color color;
        float position;

        public GradientPoint(Color color, float position)
        {
            this.color = color;
            this.position = position;
        }

        public Color Color
        {
            get { return color; }
        }

        public float Position
        {
            get { return position; }
        }
    }
}
