﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;


namespace NWNHeatMap
{
    class HeatMap
    {
        private List<HeatMapPoint> hmpl;
        private float scaling;
        private float dotradius;
        private float opacity;
        private bool grayscale;
        private bool trimheatless;
        private Gradient gradient;
        private Image area;

        /// <summary>
        /// Creates a new HeatMap.
        /// </summary>
        /// <param name="bgimage"></param>
        /// <param name="scaling">Number of pixels per world coordinate. 1.6f for original NWN minimap size (16 pixels per 10 meters).</param>
        /// <param name="dotradius"></param>
        /// <param name="opacity"></param>
        /// <param name="grayscale"></param>
        /// <param name="trimheatless"></param>
        public HeatMap(Image bgimage, float scaling, float dotradius, float opacity, bool grayscale, bool trimheatless)
        {
            hmpl = new List<HeatMapPoint>();
            this.scaling = scaling;
            this.dotradius = dotradius;
            this.opacity = opacity;
            this.grayscale = grayscale;
            this.trimheatless = trimheatless;

            // Load the area background image
            this.area = bgimage;
        }

        public void AddPoint(float x, float y)
        {
            // Convert coords to screen coords and add
            float pt_sx = x * scaling;
            float pt_sy = area.Height - (y * scaling);
            hmpl.Add(new HeatMapPoint(pt_sx, pt_sy));
        }

        public Bitmap GetBitmap()
        {
            // This pass calculates the pixel intensity contribution from each point and sums them.
            float[,] pix_intensity = new float[area.Width, area.Height];

            int h = area.Height;
            int w = area.Width;
            foreach (HeatMapPoint pt in hmpl)
            {
                // This approach may introduce aliasing; employing some antialiasing technique would look slightly nicer.
                int l = (int)Math.Max(pt.X - dotradius, 0);
                int r = (int)Math.Min(pt.X + dotradius, w - 1);
                int t = (int)Math.Max(pt.Y - dotradius, 0);
                int b = (int)Math.Min(pt.Y + dotradius, h - 1);
                for (int x = l; x <= r; x++)
                {
                    for (int y = t; y <= b; y++)
                    {
                        pix_intensity[x, y] = pix_intensity[x, y] + PixelIntensity(pt, x, y);
                    }
                }
            }

            List<float> intensities = new List<float>();
            for (int x = 0; x < w; x++)
            {
                for (int y = 0; y < h; y++)
                {
                    intensities.Add(pix_intensity[x, y]);
                }
            }
            intensities.Sort();
            gradient = GetGradientFromIntensities(intensities);
            float min_intensity = intensities[0];
            float max_intensity = intensities[intensities.Count - 1];

            // This pass normalizes the intensities to fall between 0.0 and 1.0.
            for (int x = 0; x < area.Width; x++)
            {
                for (int y = 0; y < area.Height; y++)
                {
                    pix_intensity[x, y] = Map1DSpace(pix_intensity[x, y], min_intensity, max_intensity, 0.0f, 1.0f);
                }
            }

            // This pass constructs the output image.
            Bitmap bmp = new Bitmap(area);

            for (int x = 0; x < area.Width; x++)
            {
                for (int y = 0; y < area.Height; y++)
                {
                    Color col = bmp.GetPixel(x, y);
                    if (grayscale)
                    {
                        float pixlum = col.GetBrightness();
                        col = Color.FromArgb((int)(pixlum * 255.0f), (int)(pixlum * 255.0f), (int)(pixlum * 255.0f));
                    }
                    if (trimheatless && pix_intensity[x, y] == 0.0f)
                        bmp.SetPixel(x, y, col);
                    else
                    {
                        Color heatcol = IntensityToColor(pix_intensity[x, y]);
                        Color newcol = HeatMap.Lerp(col, heatcol, opacity);
                        bmp.SetPixel(x, y, newcol);
                    }
                }
            }

            // Return the finished image
            return bmp;
        }

        private Gradient GetGradientFromIntensities(List<float> intensities)
        {
            List<float> divs = Gradient.GetDivisionsFromData(intensities, 4);
            Gradient g = new Gradient(Color.FromArgb(64, 64, 64), Color.FromArgb(255, 0, 0)); // gray to red
            g.AddPoint(Color.FromArgb(0, 0, 128), divs[0]); // blue
            g.AddPoint(Color.FromArgb(0, 160, 158), divs[1]); // cyan
            g.AddPoint(Color.FromArgb(0, 192, 0), divs[2]); // green
            g.AddPoint(Color.FromArgb(221, 224, 0), divs[3]); // yellow

            return g;
        }

        /// <summary>
        /// Maps a value in one 1D range proportionally to another range.
        /// </summary>
        /// <param name="orig">Original value</param>
        /// <param name="a">Bottom of original range</param>
        /// <param name="b">Top of original range</param>
        /// <param name="c">Bottom of new range</param>
        /// <param name="d">Top of new range</param>
        /// <returns>Newly mapped value</returns>
        private float Map1DSpace(float orig, float a, float b, float c, float d)
        {
            return (orig - a) * (d - c) / (b - a) + c;
        }

        private Color IntensityToColor(float p)
        {
            return gradient.MapPosToColor(p);
        }

        private float PixelIntensity(HeatMapPoint pt, int x, int y)
        {
            // Add 0.5f to each pixel to place the calculation in the middle of the pixel rather than the top left corner.
            float fx = (float)x + 0.5f;
            float fy = (float)y + 0.5f;
            float distance = (float)Math.Sqrt(Math.Pow(x - pt.X, 2.0) + Math.Pow(y - pt.Y, 2.0));

            // Now that we have distance, we can apply the intensity function.
            // Current intensity function is linear falloff from 1.0 at dist=0 to 0.0 at dist>=dotradius.
            // Thought: could have points with an intensity of less than 1.0 (but >0), so you can do fun things like fade out old data over time
            if (distance >= dotradius) return 0.0f;
            else return 1.0f - (distance / dotradius);
        }

        /// <summary>
        /// Lerp between two colors
        /// </summary>
        /// <param name="c1">First color to interpolate between</param>
        /// <param name="c2">Second color to interpolate between</param>
        /// <param name="weight">Value between 0.0 and 1.0 indicating how much to weight c2 over c1</param>
        /// <returns>Interpolated color</returns>
        public static Color Lerp(Color c1, Color c2, float weight)
        {
            float r1 = (float)c1.R / 255.0f;
            float g1 = (float)c1.G / 255.0f;
            float b1 = (float)c1.B / 255.0f;
            float r2 = (float)c2.R / 255.0f;
            float g2 = (float)c2.G / 255.0f;
            float b2 = (float)c2.B / 255.0f;
            return Color.FromArgb((int)(((weight * r2) + (1.0f - weight) * (r1)) * 255.0),
                (int)(((weight * g2) + (1.0f - weight) * (g1)) * 255.0),
                (int)(((weight * b2) + (1.0f - weight) * (b1)) * 255.0));
        }
    }
}
